//! The `Classifier` manages headings.
//!
//! # Example
//! ```
//! use risp::lawbuilder::classifier::Classifier;
//! use risp::lawbuilder::classifier_default::contains;
//!
//! let classifier = Classifier::new("Theilstück", std::sync::Arc::new(contains));
//!
//! assert!(classifier.is_responsible_for("1. Theilstück"));
//! assert!(!classifier.is_responsible_for("1. Unterstück"));
//! ```

use std::{
    cell::RefCell,
    fmt::{self, Display},
    rc::Rc,
    sync::Arc,
};

use serde::{Deserialize, Serialize};

pub struct Classifier {
    pub(crate) name: String, // Hauptstück, Theil, Abschnitt, ol
    is_responsible_for: Arc<dyn Fn(&Self, &str) -> bool>,
    instances: Vec<ClassifierInstance>,
    child: Vec<Rc<RefCell<Classifier>>>,
    pub(crate) root: bool,
}

impl PartialEq for Classifier {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

impl Classifier {
    pub fn new(name: &str, is_responsible_for: Arc<dyn Fn(&Self, &str) -> bool>) -> Self {
        Self {
            name: name.into(),
            is_responsible_for,
            child: Vec::new(),
            instances: Vec::new(),
            root: false,
        }
    }

    pub fn root(self) -> Self {
        Self { root: true, ..self }
    }

    pub fn is_responsible_for(&self, name: &str) -> bool {
        (self.is_responsible_for)(&self, name)
    }
}

#[derive(Clone, PartialEq)]
pub(crate) struct ClassifierInstance {
    pub(crate) name: String, //e.g. 1 Theilstück
    pub(crate) desc: Option<String>,
    pub(crate) sections: Vec<Section>,
    pub(crate) children: Vec<Rc<RefCell<ClassifierInstance>>>,
    pub(crate) parent: Option<Rc<RefCell<ClassifierInstance>>>,
}

impl ClassifierInstance {
    fn new(name: &str) -> Self {
        Self {
            name: name.into(),
            desc: None,
            parent: None,
            sections: Vec::new(),
            children: Vec::new(),
        }
    }

    pub(crate) fn set_parent(&mut self, parent: Rc<RefCell<ClassifierInstance>>) {
        self.parent = Some(parent);
    }

    pub(crate) fn get_parent(&self) -> Option<Rc<RefCell<ClassifierInstance>>> {
        self.parent.clone()
    }

    pub(crate) fn add_child(&mut self, child: Rc<RefCell<ClassifierInstance>>) {
        self.children.push(child);
    }

    pub(crate) fn set_desc(&mut self, desc: &str) {
        self.desc = Some(desc.into());
    }

    pub(crate) fn add_section(&mut self, section: Section) {
        self.sections.push(section);
    }
}

impl std::fmt::Debug for ClassifierInstance {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Classifier")
            .field("name", &self.name)
            .field("desc", &self.desc)
            .field("sections", &self.sections)
            .field("children", &self.children)
            .finish()
    }
}

impl From<&str> for ClassifierInstance {
    fn from(value: &str) -> Self {
        Self::new(value)
    }
}

#[derive(Clone, PartialEq, Serialize, Deserialize)]
pub struct Section {
    pub(crate) symb: String, // §"1", §"2", ...
    pub(crate) par_header: Option<String>,
    pub(crate) content: Content,
}

impl fmt::Debug for Section {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let par_header = self.par_header.as_deref().unwrap_or("");
        write!(f, "{} ({})", self.symb, par_header)
    }
}

impl fmt::Display for Section {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(header) = &self.par_header {
            f.write_str(&format!("{} ({})\n{}", self.symb, header, self.content))
        } else {
            f.write_str(&format!("{}\n{}", self.symb, self.content))
        }
    }
}

#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub enum Content {
    Text(String),       //This is my direct law text
    Item(Vec<Content>), //(1) This is general law. (2) This is more specific law
    List(Vec<Content>),
}

impl Display for Content {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Text(a) => f.write_str(&format!("{a}\n")),
            Self::Item(a) => {
                let mut ret = String::new();
                for aa in a {
                    ret.push_str(&format!("{aa}"));
                }
                f.write_str(&ret)
            }
            Self::List(a) => {
                let mut ret = String::new();
                for aa in a {
                    ret.push_str(&format!("{aa}"));
                }
                f.write_str(&ret)
            }
        }
    }
}
