//! The `LawBuilder` structures law texts in its (sub-)headings.
//!
//! # Example
//! ```
//! use risp::lawbuilder::{Law, LawBuilder};
//! use risp::lawbuilder::classifier::{Classifier, Content};
//! use risp::lawbuilder::classifier_default::contains;
//! use std::sync::Arc;
//!
//! let mut builder = LawBuilder::new();
//!
//! builder.new_classifier(Classifier::new("Theilstück", Arc::new(contains)));
//!
//! builder.new_header("1. Theilstück".into());
//! builder.new_desc("Desc of 1. Theilstück".into());
//! builder.next_par_header("Topic of next paragraph".into());
//! builder.new_par("Par 1:".into(), Content::Text("Test".into()));
//!
//! let law: Law = builder.into();
//! assert_eq!(format!("{law:?}"), r#"1. Theilstück (Desc of 1. Theilstück)
//! Topic of next paragraph
//! Par 1: Test"#);
//! ```

use std::{cell::RefCell, rc::Rc};

use log::{debug, info};

use crate::lawbuilder::classifier::{ClassifierInstance, Section};

use self::classifier::{Classifier, Content};

pub mod classifier;
pub mod classifier_default;

pub struct LawBuilder {
    classifiers: Vec<Classifier>,

    /// Instances
    pub(crate) header: Vec<Rc<RefCell<ClassifierInstance>>>,

    last_instance: Option<Rc<RefCell<ClassifierInstance>>>,

    next_para_header: Option<String>,
}

impl LawBuilder {
    pub fn new() -> Self {
        Self {
            classifiers: Vec::new(),
            last_instance: None,
            header: Vec::new(),
            next_para_header: None,
        }
    }

    pub fn new_classifier(&mut self, classifier: Classifier) {
        self.classifiers.push(classifier);
    }

    pub fn new_header(&mut self, name: String) {
        let name = name.trim();

        info!("new_header={name}");

        let responsible_class = self
            .responsible_classifier(name)
            .expect(&format!("No classifier for '{name}'"));

        let mut heading: ClassifierInstance = name.into();

        if let Some(last_instance) = &self.last_instance {
            let cur = Some(last_instance.clone());

            let parent = &self.find_parent(cur, responsible_class);
            match parent {
                None => {
                    if responsible_class.root {
                        let c = Rc::new(RefCell::new(heading));
                        self.header.push(c.clone());
                        self.last_instance = Some(c.clone());
                    } else {
                        heading.set_parent(last_instance.clone());
                        let c = Rc::new(RefCell::new(heading));
                        last_instance.borrow_mut().add_child(c.clone());
                        self.last_instance = Some(c.clone());
                    }
                }
                Some(parent) => {
                    heading.set_parent(parent.clone());
                    let c = Rc::new(RefCell::new(heading));
                    parent.borrow_mut().add_child(c.clone());
                    self.last_instance = Some(c.clone());
                }
            }
        } else {
            let c = Rc::new(RefCell::new(heading));
            self.header.push(c.clone());
            self.last_instance = Some(c.clone());
        }
    }

    pub fn new_desc(&self, desc: String) {
        let desc = desc.trim();

        debug!("new_desc={desc}");
        self.last_instance
            .clone()
            .unwrap()
            .borrow_mut()
            .set_desc(desc);
    }
    pub fn next_par_header(&mut self, header: String) {
        if let Some(next_para_header) = &self.next_para_header {
            self.new_header(next_para_header.clone()); // promote to bigger header :-)
        }

        debug!("new_next_para_header={header}");
        self.next_para_header = Some(header.trim().into());
    }

    pub fn new_par(&mut self, par: String, content: Content) {
        debug!("new_par=par:{par};content:{content:#?}");

        let par_header = self.next_para_header.clone();
        self.next_para_header = None;

        self.last_instance
            .clone()
            .expect("Expect at least one classifier")
            .borrow_mut()
            .add_section(Section {
                symb: par,
                par_header,
                content,
            })
    }

    fn responsible_classifier(&self, header: &str) -> Option<&Classifier> {
        self.classifiers
            .iter()
            .find(|&c| c.is_responsible_for(header))
    }

    fn find_parent(
        &self,
        cur: Option<Rc<RefCell<ClassifierInstance>>>,
        class: &Classifier,
    ) -> Option<Rc<RefCell<ClassifierInstance>>> {
        let mut cur = cur;
        while let Some(c) = cur {
            let (cur_name, cur_parent) = {
                let c_borrow = c.borrow();
                (c_borrow.name.clone(), c_borrow.parent.clone())
            };

            let cur_responsible_class = self.responsible_classifier(&cur_name).unwrap();
            if cur_responsible_class == class {
                return c.borrow_mut().get_parent();
            }

            cur = cur_parent;
        }
        None
    }
}

#[derive(Debug)]
pub struct Law {}

impl From<LawBuilder> for Law {
    fn from(builder: LawBuilder) -> Self {
        Self {}
    }
}
